#!/usr/bin/perl -w

###########################################################################
#
# translator.pl
#
# A component of the Greenstone digital library software
# from the New Zealand Digital Library Project at the 
# University of Waikato, New Zealand.
#
# Copyright (C) 1999 New Zealand Digital Library Project
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
###########################################################################


BEGIN {
    die "GSDLHOME not set\n" unless defined $ENV{'GSDLHOME'};
    unshift (@INC, "$ENV{'GSDLHOME'}/perllib");
}

use parsargv;
use util;
use Cwd;
use File::Basename;
use GDBM_File;
use CGI;

sub main
{
    #gets the name of the language that is being translated from base language
    my $file = shift(@_);
    
    my $baselanguage = shift(@_);

    #checks that argument was supplied
    if (!$file) {
	die "You didn't supply the name of the language file! Aborting...\n";
    }
    $file = "port" if ($file =~ m/portuguese/);
    $file = "indo" if ($file =~ m/indonesian/);

    #saves the directory path to the database files
    my $gdbmdir  = "$ENV{'GSDLHOME'}/tmp/lang/gdbmfiles/";
    #creates the directory if it doesn't exist
    if (!-e $gdbmdir) {
	my $store_umask = umask(0000);
	if (! mkdir($gdbmdir, 0777)) {
	    umask($store_umask);
	    die "Couldn't create directory $gdbmdir\n";
	}
	umask($store_umask);
    }
    #saves the directory path to the macro files
    my $macrodir = "$ENV{'GSDLHOME'}/macros/";
    
     # always do this update because english.dm changes often
    #hashes the macro file into form of macroname{macrotext}
    my ($englhash) = &parse_macrofile($macrodir.$baselanguage.".dm");
    my ($engl2hash) =  &parse_macrofile($macrodir.$baselanguage."2.dm");
    $englhash = &combine_hashes($englhash,$engl2hash);

    &file_to_db($baselanguage, $englhash);
    
    my ($base) = &dateannotation($baselanguage);
    my ($base2) = &dateannotation($baselanguage."2");
    $base = &combine_hashes($base,$base2);

    # language macro-filename and database-filename
    my $macrofile = $macrodir.$file.".dm";
    my $dbfile = $gdbmdir.$file.".db";
    
    if (! -e $macrofile) {
	
	@packages = ("Global", "about", "document", "query", "preferences", "browse", "help",
		     "home", "homehelp", "extlink", "authen", "collector", "docs", "usersedituser",
		     "usersdeleteuser", "bsummary", "status", "users", "gsdl", "userschangepasswd",
		     "userschangepasswdok");

	open MCROFILE, ">$macrofile" or die "MURGH\n";
	
	foreach $packg (@packages) {
	    print MCROFILE "package $packg\n";
	    print MCROFILE "# text macros\n";
	    print MCROFILE "# icons\n";
	}

	close MCROFILE;

	$code = $& if($file =~ m/\w\w/);

	open LANGFILE, "<$ENV{'GSDLHOME'}/tmp/lang/package_forms/languages.log" or die "MURGH\n";

	while (<LANGFILE>) {
	    if ($_ eq $code){ 
		$code = $& if($file =~ m/.\w\w/);
		$code =~ s/\A.//;
	    }
	}

	close LANGFILE;
	
	open LANGFILE, ">>$ENV{'GSDLHOME'}/tmp/lang/package_forms/languages.log" or die "MURGH\n";
	print LANGFILE "\n$file";
	$code = $& if($file =~ m/\w\w/);
	print LANGFILE "\n$code";
	close LANGFILE;
    }
   
    # if it doesn't have a database file then create one afresh
    my ($langhash) = &parse_macrofile($macrofile);
    $macrofile = $macrodir.$file."2.dm";
    my ($langhash2) = &parse_macrofile($macrofile);
    $langhash = &combine_hashes($langhash, $langhash2);

    &file_to_db($file, $langhash);
    
    my $foreign = &dateannotation($file);
    my $foreign2 = &dateannotation($file . "2");
    $foreign = &combine_hashes($foreign, $foreign2);

    #finds the differences between the english database (base language)
    #and the supplied languages database and hashes these differences
    #in the form macroname{macrotext}
    my $diffhash = &db_db_diffs($gdbmdir.$baselanguage.".db", $dbfile, $base, $foreign, $baselanguage);
    #generates HTML code to display user interface on webpage with what needs translation
    #returns an array of CGI data
   
    $file = "indonesian" if ($file =~ m/indo/);
    $file = "portuguese" if ($file =~ m/port/);

    my @queries = &generate_pages($diffhash, $file, $baselanguage);

}

sub combine_hashes {
    
    my ($hash1, $hash2) = @_;
    my %combined = ();

    foreach $key (keys %$hash1) {
	$combined{$key} = $hash1->{$key};
    }
    foreach $key (keys %$hash2) {
	$combined{$key} = $hash2->{$key};
    }

    return (\%combined);

}

sub parse_macrofile
{
    #saves the path of the macro file
    my $filename = shift(@_);
    #initialises some local variables
    my $currpackage;
    my %macros = ();
    my @date = localtime(time);
    
    #opens macro file or kills program
    open(IN, "<$filename") or return;
    #reads in contents of macro file, line by line
    while (<IN>) {
	# find out the current package that macros belong to
	if (s/^package //) {
	    $currpackage = $_;
	    chomp($currpackage);
	}
	#if line contains a macro name
	elsif (/^(_\w+_)/) {
	    # gets the name of the macro ($1 contains text matched by corresponding
	    #set of parentheses in the last matched pattern within the dynamic scope
	    #which here matches (_\w+_) )
	    my $macroname = $1;
	    #saves the day/month/year
	    my $macrotext = $date[3]. "/" . $date[4] . "/" . $date[5] ."\n";
	    # key to the hash and the database
	    my $key = $currpackage . "::" . $macroname;
	    # get the first line of the macro
	    $_ =~ s/^_\w+_ *//;
	    # while there is still text of the macro to go...
	    while ($_ !~ /.*\}/) {
		# ... adds it to the macrotext variable
		$macrotext .= $_;
		$_ = <IN>;
	    }
	    $macrotext .= $_;
	    #hashes the macroname and the macrotext
	    $macros{$key} = $macrotext;
	}
	#line in format ## "sometext" ## macro ## macroname ##
	elsif (/^\#\# .*/) {
	    my $macroname = $_;
	    #saves the day/month/year
	    my $macrotext = $date[3]. "/" . $date[4] . "/" . $date[5] ."\n";
	    	   
	    unless ($macroname =~ m/^\#\# .*\#\#/) {
		
		$macroname = $_;
		chomp($macroname);
		$_ = <IN>;
		$macroname .= $_;
		
	    }

	    my @names = split(/\s*\#\#\s*/, $macroname);
	    
	    $macrotext .= $macroname;
	    $_ = <IN>;

	    $macroname = $names[(scalar @names) - 1];
	    
	    my $key = $currpackage . "::" . $macroname; # key to the hash and the database

	    while ($_ !~ /^\s+/) {
		$macrotext .= $_;
		$_ = <IN>;
	    }

	    $macros{$key} = $macrotext;
	}
	
    }    
    close IN;
    
    return (\%macros);
}

sub dateannotation {
    
    $language = shift(@_);
    my %macrodated = ();

    $file = qx/cd $ENV{'GSDLHOME'}\/macros;cvs annotate $language.dm/;
    @file = split(/\n/,$file);
   
    for ($p = 0; $p < scalar(@file); $p++) {
	
	$line = $file[$p];
	chomp($line);
	$line =~ s/.*\(.*\):\s//;
	$date = $&;
	
	$date = $& if($date =~ m/\d\d-\w\w\w-\d\d/);
	
	if ($line =~ m/\#\# .*/) {
	    	   
	    unless ($line =~ m/\A\#\# .*\#\#/) {
		$line = $file[++$p];
		chomp($line);
		$line =~ s/.*\(.*\):\s//;
	    }
	    my @icon = split(/\s*\#\#\s*/, $line);
	    
	    $macro = $icon[(scalar @icon) - 1];
	    
	    $macrodated{$macro} = $date;
	    
	    while ($line =~ m/\s+/) {
		$line = $file[++$p];
		chomp($line);
		$line =~ s/.*\(.*\):\s//;
	    }
	}
	elsif ($line =~ m/\A(_\w+_)/) {
	    $line = $&;
	    $macrodated{$line} = $date;
	}
    }
    
    return (\%macrodated);
}


#######################
# DATABASE OPERATIONS #
#######################

# At the moment this routine is used to update english.db
# file when there are change in english.dm macro file; it
# has been made generic so that other database files can be
# updated if the need arises in the future.
sub file_to_db
{
    my ($filename, $filehash) = @_;

    my $database = "$ENV{'GSDLHOME'}/tmp/lang/gdbmfiles/$filename.db";
    my %dbhash = ();

    
    if (-e $database) {
	
	tie(%dbhash, 'GDBM_File', $database, 1, 0640) or die "$!";

	foreach $filekey (sort(keys(%$filehash))) {
	    if (!($dbhash{$filekey})) {
		$dbhash{$filekey} = $filehash->{$filekey};
	    }
	    else {
		# strip off the dates from both entries...
		my $filemacro = ($filehash->{$filekey});
		my $dbmacro = ($dbhash{$filekey});
		$filemacro =~ s/^.*//;
		$dbmacro =~ s/^.*//;

		if ($dbmacro ne $filemacro) {
		    #update the database to reflect the change...
		    $dbhash{$filekey} = $filehash->{$filekey};
		}
	    }
	}
    }
    else { # create the database file for the particular language file...
	# tie the hash to the database
	tie(%dbhash, "GDBM_File", $database, GDBM_WRCREAT, 0640);
	
	%dbhash = %$filehash;
    }

    untie %dbhash;
}

# db_db_diffs => Takes two paramaters which are two database filenames, works out
#                the differences between the two, and returns them in %diffhash
#
# ARGUMENTS: $db_one is the database which contains all the values, e.g. english
#            $db_two is the database which is possibly missing some values
#
# PRE:  $db_one and $db_two assumed to exist
# POST: %differences holds entries that are in $db_one but not in $db_two
sub db_db_diffs
{
    my ($db_one, $db_two, $basedates, $foreigndates, $baselanguage) = @_;

    my %db1hash = ();
    my %db2hash = ();
    my %months = ("Jan",1,"Feb",2,"Mar",3,"Apr",4,"May",5,"Jun",6,
		  "Jul",7,"Aug",8,"Sep",9,"Oct",10,"Nov",11,"Dec",12);
    
    # this hash holds the database entries that are in 
    # english.db but not in $filename.db
    my %diffhash = ();

    #ties the hash to database so when you read from hash it fetches from external
    #database and when you set hash it writes to external database
    tie(%db1hash, "GDBM_File", $db_one, GDBM_READER, 0640);
    tie(%db2hash, "GDBM_File", $db_two, GDBM_READER, 0640);

    #the key is the name of the macro
    foreach $key (sort(keys(%db1hash))) {
	#if the macro isn't there at all, then it's different
	if (!$db2hash{$key}) {
	    $diffhash{$key} = [$db1hash{$key},"",""];
	}
	#else if the macro is in both, want to compare date stamps of date hashes HERE
	#and if foreign date is older than base date want to add to diffhash aswell
	else {
	    $macro = $key;
	    $macro =~ s/\A.*:://;
	    $bdate = $basedates->{$macro};
	    $fdate = $foreigndates->{$macro};
	    @bdate = split(/-/,$bdate);
	    @fdate = split(/-/,$fdate);
	    if ($bdate[2] >= $fdate[2]) {
		if ($months{$bdate[1]} >= $months{$fdate[1]}) {
		    if ($bdate[0] > $fdate[0]) {
			$foreigntext = $db2hash{$key};
			$foreigntext =~ s/^.*\n//;
			$foreigntext =~ s/.*\{//;
			$foreigntext =~ s/\}\Z//;

			$file = qx/cd $ENV{'GSDLHOME'}\/macros;cvs diff -D $fdate $baselanguage.dm/;
			$file .= qx/cd $ENV{'GSDLHOME'}\/macros;cvs diff -D $fdate $baselanguage"2.dm"/;
			@file = split(/\n/,$file);

			$found = 1;
			foreach $key (@file) {
			    if ($key =~ m/$macro/) {
				if ($key =~ m/\A</) {
				    $key =~ s/\A<\s*(_\w+_)\s*\{//;
				    $key =~ s/\}\Z//;
				    print STDERR "KEY>>>>>||||| $key\n";
				}
				$found = 0;
			    }
			}
			print STDERR "NOTFOUND $macro\n" if ($found);
		    
			$diffhash{$key} = [$db1hash{$key}, $foreigntext,""];
		    }
		}
	    }
	}
    }

    untie %db1hash;
    untie %db2hash;

    return \%diffhash;    
}


#############################################
# WEB-FORM GENERATION AND ARGUMENT FETCHING #
#############################################

# generates web-form using the datahash you pass in;
# it will be broken down into pages, 15 macros per page
sub generate_pages
{

    my ($datahash) = shift(@_);
    my $lang = shift(@_);
    my $baselanguage = shift(@_);
    my %pagehash = ();
    my @query_pages = ();

    # directory in which to put the generated html pages
    my $dir = "$ENV{'GSDLHOME'}/tmp/lang/package_forms";
   
    # creates directory if it doesn't already exist
    if (!-e $dir) {
	my $store_umask = umask(0000);
	if (! mkdir($dir, 0777)) {
	    umask($store_umask);
	    die "Couldn't create directory $dir\n";
	}
	umask($store_umask);
    }
    
    @pageno = ();
    $pageno = 1;
    $keysperpage = 0;
    $limit = 15;

    # goes through hash of differences between base language and translation
    foreach $key (sort(keys(%$datahash))) {
	#in the hash of HTML pages, add this key to the current page and initialise it
	#with it's value in the datahash
	$pagehash->{$pageno}->{$key} = $datahash->{$key};
	$keysperpage++;
	
	$limit = 7 if($key =~ m/collector::_text.*/);
	$limit = 7 if(!($key =~ m/_.*\_/g));
	$limit = 7 if($key =~ m/help::_text.*/);
	$limit = 3 if($key =~ m/home::.*/);
	$limit = 3 if($key =~ m/gsdl::.*/);
	$limit = 1 if($key =~ m/help::_(.*)texthelp.*/);
	$limit = $keysperpage if ($limit <= $keysperpage);

	#if have enough keys to generate a page with
	if ($keysperpage == $limit) {
	    #resets key counting variable
	    $keysperpage = 0;
	    $limit = 15;
	    #opens a file to write the HTML code to for current page
	    open HTMLFILE, ">$dir/$pageno.lang";

	    #passes hash of keys for current page and filehandle, returns CGI data
	    my $query = &generate_form($pagehash->{$pageno}, \*HTMLFILE, $lang, $baselanguage);
	    close HTMLFILE;

	    #push the page number onto the array of pages
	    push(@pageno, $pageno);
	    #increment the current page
	    $pageno++;
	}
    }
    
    #if you exit the loop with keys still to write
    if ($keysperpage != 0) {
	
	#writes out the remaining keys, same format as above
	open HTMLFILE, ">$dir/$pageno.lang";
	my $query = &generate_form($pagehash->{$pageno}, \*HTMLFILE, $lang, $baselanguage);
	close HTMLFILE;
	push(@pageno, $pageno);
    }

    #writes each page number to a file
    open PAGELOG, ">$ENV{'GSDLHOME'}/tmp/lang/package_forms/pageno.log" or die("MURGH\n");
    foreach $page (@pageno) {
	print PAGELOG $page, "\n";
    }
    close PAGELOG;

    #write thankyou page for language translator once translation is complete
    open THANKYOU, ">$ENV{'GSDLHOME'}/tmp/lang/package_forms/thankyou.lang" or die("MURGH\n");
    print THANKYOU "<center>Thankyou for completing the $lang translation<p></center>\n";
    close THANKYOU;

    #returns CGI data
    return @queries;
}

sub generate_form
{
    my $formhash = shift(@_);
    my $fh = shift(@_);
    my $language = shift(@_);
    my $baselanguage = shift(@_);
    # common gateway interface for writing the stuff to the web
    my $query = new CGI;
    my $keynamecount = 1;

    print $fh ("<center>\n",
	       "If an<font color=\"FF0000\"> UPDATED! </font>tag is present a translation already exists,<br>\n",
	       "however it's macro has since been updated.<br>\n",
	       "Please ensure that the current translation provided in the input field is correct.<br>\n",
	       $query->hidden(-name=>"hiddenlanguage",-default=>"$language"),
	       $query->hidden(-name=>"baselanguage",-default=>"$baselanguage"),
	       "<p><table border=\"5\" cellpadding=\"10\">\n",
	       "<tr><td><strong>BASE LANGUAGE FILE<br>",
	       "________________________________</strong></td>\n",
	       "<td><strong>TRANSLATED LANGUAGE<br>",
	       "________________________________</strong></td></tr>\n");
    
    
    foreach $key (sort(keys(%$formhash))) {

	my $text = $formhash->{$key}[0];
	
	# whole lot of formatting on strings
	# escape all the '_' with '\' so that Greenstone doesn't substitute
	# the macro names for the contents of the macro itself
	my $keyname = $key;
	  
	$keyname  =~ s/_/\\_/g;
	# strip off the date
	$text =~ s/.*//;
	$text =~ s/(\[l=.*\])//;
	# tells you the date the english file was last updated...
	#$date = $&;
	
	print STDERR "$text\n";

	
	# must take care of image macros single and multiple line ones
	$text =~ s/(\<br\>\#\#.*\#\#\s*\<br\>).*/$1/g;
	
	# format the text for displaying in the browser
	$text =~ s/[\{\}]//g unless($text =~ m/\#\# /);
	$text =~ s/(\A\n)//;
	$text =~ s/(\n\Z)//;
	$text =~ s/\n/\<br\>/g;
	
	$text =~ s/</\&lt;/g;
	$text =~ s/>/\&gt;/g;
	$text =~ s/\b_([A-Za-z0-9:]+)_\b/ \b\\_$1\\_ \b/g;
	
	$keynamecount = 1;
	
	my $default = $formhash->{$key}[1];
	my $updated = "";
	$updated = "<font color=\"FF0000\"> UPDATED! </font><br>" if ($default =~ m/\S+/);

        # so that when it is an image only get the text to translate not coding stuff aswell
	if ($text =~ m/\#\# (^|.)*/) {
	    if ($& =~ m/".*"/) {
		print $fh ("<tr><td> $updated $&</td><td>",
			   $query->hidden(-name=>"$keyname" . "$keynamecount",-default=>"$`"));
		$keynamecount++;
		print $fh ($query->textfield(-name=>"$keyname" . "$keynamecount",
					     -default=>"$default",
					     -size=>50));
		$keynamecount++;
		print $fh ($query->hidden(-name=>"$keyname" . "$keynamecount",-default=>"$'"),
			   "</td></tr>\n");
	    }
	}
	elsif ($text =~ m/\S+/) {
		$text =~ s/\s+/ /g;
		$text =~ s/\&lt;br\&gt;/\&lt;br\&gt;\n/g;

		my @words = split(/ /, $text);
		my $words = scalar(@words);
		my $rows = sprintf("%.0f", $words/5);
		print $fh ("<tr><td> $updated $text </td><td>\n");
		if ($rows > 1) {
		    print $fh ($query->textarea(-name=>"$keyname" . "$keynamecount",
						-rows=>$rows * 2,
						-default=>"$default",
						-columns=>50), 
			       "<br></td></tr>\n");  
		}
		else {
		    print $fh ($query->textfield(-name=>"$keyname" . "$keynamecount",
						 -default=>"$default",
						 -size=>50),
			       "<br></td></tr>\n");
		}
	    }
    }
    #finishes table and adds a SUBMIT CHANGES option on the end
    print $fh ("</table>\n", "<br>",
	       "Please ensure your $language translation is correct before proceeding.<br>\n",
	       "Input fields may be left empty if necessary and only those field which<br>\n",
	       "contain a translation will be submitted to the macro file.<br>\n",
	       "Submitting this page will result in the next page being loaded automatically.<br><strong>\n",
	       $query->submit($pageno, 'SUBMIT TRANSLATION >>'),
	       "</strong>","</center>\n");

    # adds the CGI area just written to the array of CGI areas
    push (@queries, $query);
    
    return $query;
}

&main(@ARGV);

